import pandas as pd
import numpy as np
import argparse
from io import StringIO
import configparser
import os.path
import sys
from scipy.spatial import distance
from error_control import possible_distances

DTW_DESC_MSG = \
"""
    Args:
        -x   Time series 1
        -y   Time series 2
        -t   Calculation type DTW (str)
        -d   Type of distance (function or gower)
        -ce  Check data for errors (bool)
        -of Output to File (bool)
        -nf Name of file (str)
        -n Numero de hilos empleados para la paralelización (int)
        -k Transformación de la matriz de distancia a kernel (bool)
        -s Valor de sigma para la transformación a kernel exponencial aplicada (float)

    Others args:
        MTS  Bool value
        get_visualization  Bool value
    
    Optional arguments:
        -h, --help            show this help message and exit
        -v, --version         show version
        -l, --list            show available backends
    \n
"""

DTW_USAGE_MSG = \
"""
    %(prog)s [<args>] | --help | --version | --list \n   
""" \
+ str(DTW_DESC_MSG)

    
DTW_VERSION_MSG = \
"""
    %(prog)s 0.9.4
"""
    

def read_data(fname):
    return pd.read_csv(StringIO(fname.read()), header = None, sep=";")
    

def read_npy(fname):
    return np.load(fname.name)


class Input:
    def __init__(self):
        #Input.execute_configuration(self)
        config = configparser.ConfigParser()
        here = os.path.abspath(os.path.dirname(__file__))
        path_file_init = os.path.join(here, 'configuration.ini')

        # check if the path is to a valid file
        if not os.path.isfile(path_file_init):
            raise BadConfigError # not a standard python exception

        config.read(path_file_init)
        self.check_errors = config.getboolean('DEFAULT', 'check_errors')
        self.type_dtw = config.get('DEFAULT', 'type_dtw')
        self.MTS = config.getboolean('DEFAULT', 'MTS')
        self.n_threads = config.getint('DEFAULT', 'n_threads')
        
        # If the distance introduced is not correct, the execution is terminated.
        if self.check_errors:
            test_distance = possible_distances()
            if not config.get('DEFAULT', 'distance') in test_distance:
                raise ValueError('Distance introduced not allowed or incorrect.')
             
        #self.distance = eval("distance." + config.get('DEFAULT', 'distance'))
        self.distance = config.get('DEFAULT', 'distance')
        self.visualization = config.getboolean('DEFAULT', 'visualization')
        self.output_file = config.getboolean('DEFAULT', 'output_file')
        self.name_file = config['DEFAULT']['name_file']
        self.DTW_to_kernel = config.getboolean('DEFAULT', 'DTW_to_kernel')
        self.sigma = config.getint('DEFAULT', 'sigma')


def parse_args(isEntryFile):
    input_obj = Input()

    parser = argparse.ArgumentParser(usage=DTW_USAGE_MSG,
                             description=DTW_DESC_MSG,
                             formatter_class=argparse.RawDescriptionHelpFormatter,
                             add_help=False)

    # Control input arguments by terminal
    if isEntryFile:

        parser.add_argument('X',
                            type=argparse.FileType('r'),
                            help='POST file to be analyzed.')

        parser.add_argument('Y', nargs='?',
                            type=argparse.FileType('r'),
                            help='POST file to be analyzed.')

    else:
        parser.add_argument('-x', nargs='+', type=float, help="Temporal Serie 1")
        parser.add_argument('-y', nargs='+', type=float, help="Temporal Serie 2")

   
    parser.add_argument('-t', '--type_dtw', nargs='?', default=input_obj.MTS, type=str,
                        help="d: dependient or i: independient.")
    parser.add_argument("-d", "--distance", nargs='?', default=input_obj.distance, type=str,
                        help="Use a possible distance of scipy.spatial.distance.")
    parser.add_argument("-ce", "--check_errors", nargs='?', default=input_obj.check_errors, type=str,
                        help="Control whether or not check for errors.")
    parser.add_argument("MTS", nargs='?', default=input_obj.type_dtw, type=bool,
                        help="Indicates whether the data are multivariate time series or not.")
    parser.add_argument("visualization", nargs='?', default=input_obj.visualization, type=bool,
                        help="Allows you to indicate whether to display the results or not. Only for the one-dimensional case.")
    parser.add_argument("-of", "--output_file", nargs='?', default=input_obj.output_file, type=bool,
                        help="Output by file instead of terminal.")
    parser.add_argument("-nf", "--name_file", nargs='?', default=input_obj.name_file, type=str,
                        help="Name file.")


    # In case of working with files containing N multivariate time series, we give the possibility 
    # to determine the number of threads and whether to transform the output into a kernel.
    parser.add_argument('-n', '--n_threads', nargs='?', default=input_obj.n_threads, type=int,
                        help="d: dependient or i: independient.")
    parser.add_argument("-k", "--DTW_to_kernel", nargs='?', default=input_obj.DTW_to_kernel, type=str,
                        help="Use a possible distance of scipy.spatial.distance.")
    parser.add_argument("-s", "--sigma", nargs='?', default=input_obj.sigma, type=float,
                        help="Use a possible distance of scipy.spatial.distance.")
    

    parser.add_argument('-h', '--help', action='help',
                    help=argparse.SUPPRESS)

    parser.add_argument('-v', '--version', action='version',
                    version=DTW_VERSION_MSG,
                    help=argparse.SUPPRESS)

    parser.add_argument('-g', '--debug', dest='debug',
                    action='store_true',
                    help=argparse.SUPPRESS)

    # Save de input arguments
    args = parser.parse_args()
    input_obj.check_errors = args.check_errors
    input_obj.type_dtw = args.type_dtw
    input_obj.distance = args.distance
    input_obj.n_threads = args.n_threads
    input_obj.sigma = args.sigma
    input_obj.name_file = args.name_file
    input_obj.MTS = args.MTS
    input_obj.visualization = args.visualization
    input_obj.DTW_to_kernel = args.DTW_to_kernel
    input_obj.output_file = args.output_file
    

    # If the distance introduced is not correct, the execution is terminated.
    if input_obj.check_errors:
        test_distance = possible_distances()
        if not input_obj.distance in test_distance:
            raise ValueError('Distance introduced not allowed or incorrect.')
         
    if not input_obj.distance == "gower":
       input_obj.distance = eval("distance." + input_obj.distance)


    return parser.parse_args(), input_obj
    
    
 



# File error control of input
import numpy as np
from scipy.spatial import distance
import io
from contextlib import redirect_stdout

def control_inputs(x, y, type_dtw, MTS):
    if type_dtw == "i" and not MTS:
        raise ValueError('Get independent dtw distance only valid for MTS.')

    x = np.asanyarray(x, dtype='float')
    y = np.asanyarray(y, dtype='float')


    # Irregular multivariate time series are not allowed.
    if x.ndim == y.ndim > 1 and x.shape[1] != y.shape[1]:
        raise ValueError('Irregular multivariate time series are not allowed.')

    if x.ndim == 1 and MTS:
        raise ValueError('The data entered are not Multivariate Time Series.')

    if x.ndim == y.ndim > 1 and not MTS:
        raise ValueError('Change the value of the MTS flag.')

    return x, y



# Functions to obtain the possible distances to be managed. 
def is_distance_function(func, checker):
    with io.StringIO() as buf, redirect_stdout(buf):
        help(func)
        output = buf.getvalue()

    if output.split("\n")[0].find(checker) == -1:
        return False
    else:
        return True
        
        
def possible_distances():
   # Check that the parameter introduced by terminal associated to the 
   # distance is one of the possible parameters to use.
   possible_distance = []
   for i in range(len(dir(distance))):
	   if(len(dir(distance)[i].split("_")) == 1) and not any(c.isupper() for c in dir(distance)[i]):
		   if is_distance_function("scipy.spatial.distance."+dir(distance)[i],
		    checker = "function " + dir(distance)[i] + " in scipy.spatial.distance"):
			   possible_distance.append(dir(distance)[i])

   # I add as a possible distance measurement the gower distance. Gower distance allows the 
   # calculation of distance between continuous and binary variables. 
   possible_distance.append("gower")	  
    
   return possible_distance
   



import csv
import sys
import pandas as pd
import os.path
import numpy as np

sys.path.append(os.path.dirname(os.path.realpath(__file__)))

# Created functions
from error_control import possible_distances
from utils import *
from dtw_functions import dtw, dtw_tensor_3d



def input_File(input_obj):

    args, input_obj = parse_args(True)
    data = read_data(args.X)

    if (data.shape[0] == 2) and (data.shape[0] % 2 == 0):
        input_obj.MTS = False
        input_obj.x = data.iloc[0,:].values
        input_obj.y = data.iloc[1,:].values

    elif (data.shape[0] > 3) and (data.shape[0] % 2 == 0):
        input_obj.MTS = True
        finalData = []
        index = data.shape[0]/2
        for i in range(2):
            finalData.append(data.loc[(data.shape[0]/2)*i:index-1, :].values)
            index+=int(data.shape[0] / 2)

        input_obj.x = finalData[0]
        input_obj.y = finalData[1]

    return input_obj

def control_Output(input_obj, dtw_distance):

    if input_obj.output_file:
        sys.stdout.write("Output to "  + input_obj.name_file + ".csv")
        pd.DataFrame(np.array([dtw_distance])).to_csv(input_obj.name_file + ".csv", float_format='%g', index=False)
    else:
        sys.stdout.write(str(dtw_distance))

def main():
    # If you only make use of the library you will get an error.
    if len(sys.argv) == 1:
        s = "\nIt needs input arguments.\n"
        sys.stderr.write(s)
        sys.exit(0)
	
	# Generate an object with the deafult parameters
    input_obj = Input()
	
    # Input type 1: input by files
    if os.path.exists(sys.argv[1]):
        # input 2D file
        if sys.argv[1].endswith('.csv'):
            input_obj = input_File(input_obj)
            dtw_distance = dtw(input_obj.x, input_obj.y, input_obj.type_dtw, input_obj.distance,
                               input_obj.MTS, input_obj.visualization, input_obj.check_errors)

        # input 3D file. We include the possibility to parallelise.
        elif sys.argv[1].endswith('.npy'):

            args, input_obj = parse_args(True)
            if len(sys.argv) > 2 and sys.argv[2].endswith('.npy'):
                X, Y = read_npy(args.X), read_npy(args.Y)
            else:
                X = read_npy(args.X)
                Y = X

            dtw_distance = dtw_tensor_3d(X, Y, input_obj)
            
        control_Output(input_obj, dtw_distance)
            
    # Input type 2: input by terminal
    elif sys.argv[1] == "-x":        
        args, input_obj = parse_args(False)

        input_obj.x = args.x
        input_obj.y = args.y
       
        dtw_distance = dtw(input_obj.x, input_obj.y, input_obj.type_dtw, input_obj.distance, input_obj.MTS,
                           input_obj.visualization, input_obj.check_errors)
        
        control_Output(input_obj, dtw_distance)
        
    else:
        args, input_obj = parse_args(False)
        #s = "\nError in input arguments.\n"
        #sys.stderr.write(s)
        sys.exit(0)


if __name__ == "__main__":

   try:
      main()
   except KeyboardInterrupt:
      s = "\n\nReceived Ctrl-C or other break signal. Exiting.\n"
      sys.stderr.write(s)
      sys.exit(0)






import numpy as np
from collections import defaultdict
import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sbn

import gower
from joblib import Parallel, delayed
from scipy.spatial import distance

import sys
import os.path
sys.path.append(os.path.dirname(os.path.realpath(__file__)))

from error_control import control_inputs


# Functions that allow the display of the cost matrix together with the
# path followed to obtain the DTW distance value.

# Function that allows to obtain the path, that is, the route to
# reach the DTW distance value.
def get_path(x, y, D, path):
	
    i, j = len(x), len(y)
    while not (i == j == 0):
        # Add indexes to the path
        path.append((i - 1, j - 1))
        # Update indices
        i, j = D[i, j][1], D[i, j][2]
    # Flip the path to get it in the right order.
    return path.reverse()


# Function to obtain the cost matrix. 
def get_cost_matrix(D):
    
    positions_object = list(D)
    index = 0
    length_rows, length_columns = list(D.keys())[-1][0] + 1, list(D.keys())[-1][1] + 1
    cost_matrix = np.zeros((length_rows, length_columns))
    for i in range(length_rows):
        for j in range(length_columns):
            cost_matrix[positions_object[index]] = list(D.values())[index][0]
            index += 1

    arr = np.delete(cost_matrix, np.s_[0], 1)
    return np.delete(arr, np.s_[0], 0)


# Function for the calculation of the independent DTW distance.
def dtw_ind(x, y, dist, dtw_distance=0):
	
    dim_m = x.shape[1]
    for index_m in range(dim_m):
        x_aux = x[:, index_m]
        y_aux = y[:, index_m]
        iter_object = [(1, 1) for i in range(x_aux.ndim) for j in range(y_aux.ndim)]
        # Starting conditions:
        # 1) D(0,0) = 0
        # 2) D(i,0) = inf
        # 3) D(0,j) = inf
        D = defaultdict(lambda: (float('inf'),))
        D[0, 0] = (0, 0, 0)
        if dist == "gower":
            for i, j in iter_object:
                # Given the function distance as input parameter the distance between x and y is calculated.
                # In case of using gower distance we fit the data.
                df = pd.DataFrame(np.array([x_aux, y_aux]))
                dt = gower.gower_matrix(df)[1][0]
                D[i, j] = min((D[i - 1, j][0] + dt, i - 1, j),
                              (D[i, j - 1][0] + dt, i, j - 1),
                              (D[i - 1, j - 1][0] + dt, i - 1, j - 1))
        else:
            for i, j in iter_object:
                # Given the function distance as input parameter the distance between x and y is calculated.
                # In case of using gower distance we fit the data.
                dt = dist(x_aux, y_aux)
                D[i, j] = min((D[i - 1, j][0] + dt, i - 1, j),
                              (D[i, j - 1][0] + dt, i, j - 1),
                              (D[i - 1, j - 1][0] + dt, i - 1, j - 1))

        dtw_distance += D[x_aux.ndim, y_aux.ndim][0]

    return dtw_distance
    
    
# Function to paint the cost matrix.
def plot_cost_matrix(warp_path, cost):
    fig, ax = plt.subplots(figsize=(12, 10))
    ax = sbn.heatmap(cost, annot=True, square=True, linewidths=0.1, cmap="YlGnBu", ax=ax)

    # Get the warp path in x and y directions
    path_x = [p[0] for p in warp_path]
    path_y = [p[1] for p in warp_path]

    # Align the path from the center of each cell
    path_xx = [x + 0.5 for x in path_x]
    path_yy = [y + 0.5 for y in path_y]

    ax.plot(path_xx, path_yy, color='blue', linewidth=3, alpha=0.2)


# Function for the calculation of the dependent DTW distance.
def dtw_dep(x, y, dist):
    iter_object = [(i + 1, j + 1) for i in range(len(x)) for j in range(len(y))]
    # Starting conditions:
    # 1) D(0,0) = 0
    # 2) D(i,0) = inf
    # 3) D(0,j) = inf
    D = defaultdict(lambda: (float('inf'),))
    D[0, 0] = (0, 0, 0)
    if dist == "gower":
        for i, j in iter_object:
            # Given the function distance as input parameter the distance between x and y is calculated.
            # In case of using gower distance we fit the data.
            df = pd.DataFrame(np.array([x[i - 1], y[j - 1]]))
            dt = gower.gower_matrix(df)[1][0]
            D[i, j] = min((D[i - 1, j][0] + dt, i - 1, j),
                          (D[i, j - 1][0] + dt, i, j - 1),
                          (D[i - 1, j - 1][0] + dt, i - 1, j - 1))
    else:
        for i, j in iter_object:
            # Given the function distance as input parameter the distance between x and y is calculated.
            dt = dist(x[i - 1], y[j - 1])
            D[i, j] = min((D[i - 1, j][0] + dt, i - 1, j),
                          (D[i, j - 1][0] + dt, i, j - 1),
                          (D[i - 1, j - 1][0] + dt, i - 1, j - 1))

    return D[len(x), len(y)][0], D





def dtw(x, y, type_dtw="d", dist=distance.euclidean, MTS=False, get_visualization=False, check_errors=False):

    if check_errors:
        x, y = control_inputs(x, y, type_dtw, MTS)

    if MTS:
        if type_dtw == "i":
            dtw_distance = dtw_ind(x, y, dist)
        else:
            dtw_distance, _ = dtw_dep(x, y, dist)
    else:
        dtw_distance, D = dtw_dep(x, y, dist)

    if get_visualization:
        if type_dtw != "i" and not MTS:
            path = []
            get_path(x, y, D, path)
            cost_matrix = get_cost_matrix(D)
            plot_cost_matrix(path, cost_matrix)
        else:
            raise ValueError('Display not allowed. Only univariate case.')

    return dtw_distance


# We transform the DTW matrix to an exponential kernel. 
def transform_DTW_to_kernel(data, sigma):
	
	return np.exp(-data/(2*sigma**2))
	
	
# Function to obtain the calculation of the DTW distance at a high level. Parallelization is included.
def dtw_tensor_3d(X_1, X_2, input_obj):

    dtw_matrix_train = Parallel(n_jobs=input_obj.n_threads)(
        delayed(dtw)(X_1[i], X_2[j], type_dtw=input_obj.type_dtw, dist=input_obj.distance,
                     MTS=input_obj.MTS, get_visualization=input_obj.visualization, check_errors=input_obj.check_errors)
        for i in range(X_1.shape[0]) 
        for j in range(X_2.shape[0])
    )
    
    data = np.array(dtw_matrix_train).reshape((X_1.shape[0], X_2.shape[0]))

    if input_obj.DTW_to_kernel:
        return transform_DTW_to_kernel(data, input_obj.sigma)

    return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pathlib
import codecs
import os.path
from setuptools import setup
import re

# The directory containing this file
#HERE = pathlib.Path(__file__).resolve().parent

# The text of the README file is used as a description
#README = (HERE / "README.md").read_text()


here = os.path.abspath(os.path.dirname(__file__))
readme_md = os.path.join(here, 'README.md')
version_py = os.path.join(here, 'dtwParallel', '_version.py')

# Get the package description from the README.md file
with codecs.open(readme_md, encoding='utf-8') as f:
    long_description = f.read()

with codecs.open(version_py, 'r', encoding='utf-8') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

# This call to setup() does all the work
setup(
    name="dtwParallel",
    version=version,
    description="Python implementation of Dynamic Time Warping (DTW), which allows computing the dtw distance between one-dimensional and multidimensional time series, with the possibility of visualisation (one-dimensional case) and parallelisation (multidimensional case).",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/oscarescuderoarnanz/dtwParallel",
    author="oscarescuderoarnanz",
    author_email="escuderoarnanzoscar@gmail.com",
    license="BSD 2-clauses",
    packages=[
       "dtwParallel"
    ],
    namespace_packages=[
      "dtwParallel",
    ],
    keywords="Dynamic Time Warping Parallelization Visualisation Distance",
    #classifiers=[
    #    'Intended Audience :: Developers/Science/Reasearch',
    #    'Topic :: Scientific/Engineering/Software Development',
    #    'License :: OSI Approved :: BSD License',
    #    'Programming Language :: Python :: 3'
    #],

    include_package_data = True,
    package_data = {
    # If any package contains *.ini files, include them
    '': ['*.ini'],
    },
    py_modules={
        "utils",
        "dtw_functions",
        "error_control" 
    },
    entry_points={
        'console_scripts': [
        'dtwParallel=dtwParallel.dtwParallel:main'
        ]
    }
)
